"""Utility functions to sample from the posterior predictive distributions"""

import numpy as np

from scipy import stats


def _get_mu_and_idxs(mu, draws, draw_n):  # pylint: disable=unused-argument
    """Sample values from the posterior of the mean, return sampled values and their indexes

    This function is used within ``pps_*`` auxiliary functions. Its goal is to simplify how we
    sample from the posterior trace and the other parameters that may define the conditional
    distribution of the response.

    Parameters
    ----------
    mu: np.array
        A 3-dimensional numpy array with samples from the posterior distribution of the mean. The
        first dimension represents the chain, the second represents the draw, and the third the
        individual.
    draws: int
        Number of samples to take.
    draw_n: int
        Number of draws available in per chain of the posterior

    Returns
    -------
    mu: np.array
        A 3-dimensional numpy array with the new samples.
    idxs: np.array
        The indexes of the samples in the original sample of the posterior.
    """
    # mu has shape (chain, draw, obs)
    idxs = np.random.randint(low=0, high=draw_n, size=draws)
    mu = mu[:, idxs, :]
    return mu, idxs


def pps_bernoulli(model, posterior, mu, draws, draw_n):  # pylint: disable=unused-argument
    mu, _ = _get_mu_and_idxs(mu, draws, draw_n)
    return np.random.binomial(1, mu)


def pps_beta(model, posterior, mu, draws, draw_n):
    mu, idxs = _get_mu_and_idxs(mu, draws, draw_n)
    kappa = posterior[model.response.name + "_kappa"].values[:, idxs, np.newaxis]
    alpha = mu * kappa
    beta = (1 - mu) * kappa
    return np.random.beta(alpha, beta)


def pps_binomial(
    model, posterior, mu, draws, draw_n, trials=None
):  # pylint: disable=unused-argument
    mu, _ = _get_mu_and_idxs(mu, draws, draw_n)
    if trials is None:
        trials = model.response.data[:, 1]
    return np.random.binomial(trials.squeeze(), mu)


def pps_gamma(model, posterior, mu, draws, draw_n):
    mu, idxs = _get_mu_and_idxs(mu, draws, draw_n)
    alpha = posterior[model.response.name + "_alpha"].values[:, idxs, np.newaxis]
    beta = alpha / mu
    return np.random.gamma(alpha, 1 / beta)


def pps_gaussian(model, posterior, mu, draws, draw_n):
    mu, idxs = _get_mu_and_idxs(mu, draws, draw_n)
    sigma = posterior[model.response.name + "_sigma"].values[:, idxs, np.newaxis]
    return np.random.normal(mu, sigma)


def pps_negativebinomial(model, posterior, mu, draws, draw_n):
    mu, idxs = _get_mu_and_idxs(mu, draws, draw_n)
    n = posterior[model.response.name + "_alpha"].values[:, idxs, np.newaxis]
    p = n / (mu + n)
    return np.random.negative_binomial(n, p)


def pps_poisson(model, posterior, mu, draws, draw_n):  # pylint: disable=unused-argument
    mu, _ = _get_mu_and_idxs(mu, draws, draw_n)
    return np.random.poisson(mu)


def pps_t(model, posterior, mu, draws, draw_n):
    mu, idxs = _get_mu_and_idxs(mu, draws, draw_n)
    if isinstance(model.family.likelihood.priors["nu"], (int, float)):
        nu = model.family.likelihood.priors["nu"]
    else:
        nu = posterior[model.response.name + "_nu"].values[:, idxs, np.newaxis]

    sigma = posterior[model.response.name + "_sigma"].values[:, idxs, np.newaxis]
    return stats.t.rvs(nu, mu, sigma)


def pps_wald(model, posterior, mu, draws, draw_n):
    mu, idxs = _get_mu_and_idxs(mu, draws, draw_n)
    lam = posterior[model.response.name + "_lam"].values[:, idxs, np.newaxis]
    return np.random.wald(mu, lam)



"""Classes to construct model families."""
from .family import Family, _extract_family_prior
from .likelihood import Likelihood
from .link import Link

__all__ = [
    "_extract_family_prior",
    "Family",
    "Likelihood",
    "Link",
]



FAMILY_LINKS = {
    "bernoulli": ["identity", "logit", "probit", "cloglog"],
    "beta": ["identity", "logit", "probit", "cloglog"],
    "binomial": ["identity", "logit", "probit", "cloglog"],
    "gamma": ["identity", "log", "inverse"],
    "gaussian": ["identity", "log", "inverse"],
    "negativebinomial": ["identity", "log", "cloglog"],
    "poisson": ["identity", "log"],
    "t": ["identity", "log", "inverse"],
    "wald": ["inverse", "inverse_squared", "identity", "log"],
}


def listify(obj):
    """Wrap all non-list or tuple objects in a list.

    Provides a simple way to accept flexible arguments.
    """
    if obj is None:
        return []
    else:
        return obj if isinstance(obj, (list, tuple, type(None))) else [obj]


def spacify(string):
    """Add 2 spaces to the beginning of each line in a multi-line string."""
    return "  " + "  ".join(string.splitlines(True))


def multilinify(sequence, sep=","):
    """Make a multi-line string out of a sequence of strings."""
    sep += "\n"
    return "\n" + sep.join(sequence)


def link_match_family(link, family_name):
    """Checks whether the a link can be used in a given family.

    When this function is used with built-in family names, it tests whether the link name can be
    used with the given built-in family. If the family name is not known, we return True because
    the user is working with a custom ``Family`` object.
    Which links can work with which families are taken from statsmodels.
    """
    if family_name in FAMILY_LINKS:
        return link in FAMILY_LINKS[family_name]

    # Custom family, we don't know what link functions can be used
    return True



from statsmodels.genmod import families as sm_families

from bambi.families.link import Link

STATSMODELS_FAMILIES = {
    "bernoulli": sm_families.Binomial,
    "gamma": sm_families.Gamma,
    "gaussian": sm_families.Gaussian,
    "wald": sm_families.InverseGaussian,
    "negativebinomial": sm_families.NegativeBinomial,
    "poisson": sm_families.Poisson,
}

STATSMODELS_LINKS = {
    "identity": sm_families.links.identity(),
    "logit": sm_families.links.logit(),
    "probit": sm_families.links.probit(),
    "cloglog": sm_families.links.cloglog(),
    "inverse": sm_families.links.inverse_power(),
    "inverse_squared": sm_families.links.inverse_squared(),
    "log": sm_families.links.log(),
}


class Family:
    """A specification of model family.

    Parameters
    ----------
    name : str
        The name of the family. It can be any string.
    likelihood: Likelihood
        A ``bambi.families.Likelihood`` instace specifying the model likelihood function.
    link : str or Link
        The name of the link function or a ``bambi.families.Link`` instance. The link function
        transforms the linear model prediction to the mean parameter of the likelihood funtion.

    Examples
    --------

    >>> import bambi as bmb

    Replicate the Gaussian built-in family.

    >>> sigma_prior = bmb.Prior("HalfNormal", sigma=1)
    >>> likelihood = bmb.Likelihood("Gaussian", parent="mu", sigma=sigma_prior)
    >>> family = bmb.Family("my_gaussian", likelihood, "identity")
    >>> # Then you can do
    >>> # bmb.Model("y ~ x", data, family=family)

    Replicate the Bernoulli built-in family.

    >>> likelihood = bmb.Likelihood("Bernoulli", parent="p")
    >>> family = bmb.Family("bernoulli2", likelihood, "logit")
    """

    def __init__(self, name, likelihood, link):
        self.smlink = None
        self.link = None
        self.name = name
        self.likelihood = likelihood
        self.smfamily = STATSMODELS_FAMILIES.get(name, None)
        self._set_link(link)

    def _set_link(self, link):
        """Set new link function.

        If ``link`` is of type ``str``, this method attempts to create a ``bambi.families.Link``
        from the name passed. If it is a recognized name, a builtin ``bambi.families.Link`` will be
        used. Otherwise, ``bambi.families.Link`` instantiation will raise an error.

        Parameters
        ----------
        link: str or bambi.families.Link
            If a string, it must the name of a link function recognized by Bambi.

        Returns
        -------
        None
        """
        if isinstance(link, str):
            self.link = Link(link)
            self.smlink = STATSMODELS_LINKS.get(link, None)
        elif isinstance(link, Link):
            self.link = link
        else:
            raise ValueError("'link' must be a string or a Link instance.")

    def __str__(self):
        msg_list = [f"Response distribution: {self.likelihood.name}", f"Link: {self.link.name}"]
        if self.likelihood.priors:
            priors_msg = "\n  ".join([f"{k} ~ {v}" for k, v in self.likelihood.priors.items()])
            msg_list += [f"Priors:\n  {priors_msg}"]
        msg = "\n".join(msg_list)
        return msg

    def __repr__(self):
        return self.__str__()


# Names of parameters that can receive a prior distribution for the built-in families
FAMILY_PARAMS = {
    "beta": ("kappa",),
    "gamma": ("alpha",),
    "gaussian": ("sigma",),
    "negativebinomial": ("alpha",),
    "t": ("sigma", "nu"),
    "wald": ("lam",),
}


def _extract_family_prior(family, priors):
    """Extract priors for a given family

    If a key in the priors dictionary matches the name of a nuisance parameter of the response
    distribution for the given family, this function extracts and returns the prior for that
    nuisance parameter. The result of this function can be safely used to update the ``Prior`` of
    the response term.

    Parameters
    ----------
    family: str or ``bambi.families.Family``
        The family for which we want to extract priors.
    priors: dict
        A dictionary where keys represent parameter/term names and values represent
        prior distributions.
    """
    if isinstance(family, str) and family in FAMILY_PARAMS:
        names = FAMILY_PARAMS[family]
        priors = {name: priors.pop(name) for name in names if priors.get(name) is not None}
        if priors:
            return priors
    elif isinstance(family, Family):
        # Only work if there are auxiliary parameters in the family, and if any of these are
        # present in 'priors' dictionary.
        nuisance_params = list(family.likelihood.priors)
        if set(nuisance_params).intersection(set(priors)):
            return {k: priors.pop(k) for k in nuisance_params if k in priors}
    return None



import numpy as np


class ResponseTerm:
    """Representation of a single response model term.

    Parameters
    ----------
    term: formulae.ResponseVector
        An object describing the response of the model,
        as returned by ``formulae.design_matrices().response``
    family : Family
        An ``Family`` object.
    """

    def __init__(self, term, family):
        self.name = term.name
        self.data = term.design_vector
        self.categorical = term.type == "categoric"
        self.success_event = term.success if term.success is not None else 1
        self.constant = np.var(self.data) == 0
        self.family = family

        if family.name == "bernoulli":
            if not all(np.isin(self.data, ([0, 1]))):
                raise ValueError("Numeric response must be all 0 and 1 for 'bernoulli' family.")

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"family: {self.family.name}",
            f"shape: {self.data.squeeze().shape}",
        ]
        if self.family.name == "bernoulli":
            args += [f"succes: {self.success_event}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()


class Term:
    """Representation of a single (common) model term.

    Parameters
    ----------
    name : str
        Name of the term.
    term_dict: dict
        A dictionary describing the components of the term. These can be found in
        ``formulae.design_matrices().common.terms_info``
    data : ndarray
        The term values.
    prior : Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    """

    group_specific = False

    def __init__(self, name, term_dict, data, prior=None):
        self.name = name
        self.data = data
        self.prior = prior
        self.type = term_dict["type"]
        self.levels = term_dict["full_names"]
        self.categorical = False

        # If the term has one component, it's categorical if the component is categorical.
        # If the term has more than one component (i.e. it is an interaction), it's categorical if
        # at least one of the components is categorical.
        if self.type == "interaction":
            if any((term["type"] == "categoric" for term in term_dict["terms"].values())):
                self.categorical = True
        else:
            self.categorical = self.type == "categoric"

        # Flag constant terms
        if self.categorical and len(term_dict["levels"]) == 1 and (data == data[0]).all():
            raise ValueError(f"The term '{name}' has only 1 category!")

        if not self.categorical and self.type != "intercept" and np.all(data == data[0]):
            raise ValueError(f"The term '{name}' is constant!")

        # Flag cell-means terms (i.e., full-rank coding), which receive special priors
        # To flag intercepts we use `self.type`
        self.is_cell_means = self.categorical and (self.data.sum(1) == 1).all()

        # Obtain pymc coordinates, only for categorical components of a term.
        # A categorical component can have up to two coordinates if it is including with both
        # reduced and full rank encodings.
        self.pymc_coords = {}
        if self.categorical:
            name = self.name + "_coord"
            if self.type == "interaction":
                self.pymc_coords[name] = term_dict["levels"]
            elif term_dict["encoding"] == "full":
                self.pymc_coords[name] = term_dict["levels"]
            else:
                self.pymc_coords[name] = term_dict["levels"][1:]

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"prior: {self.prior}",
            f"type: {self.type}",
            f"shape: {self.data.squeeze().shape}",
            f"categorical: {self.categorical}",
        ]
        if self.categorical:
            args += [f"levels: {self.levels}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()


class GroupSpecificTerm:
    # pylint: disable=too-many-instance-attributes
    """Representation of a single (group specific) model term.

    Parameters
    ----------
    name : str
        Name of the term.
    term: dict
        A dictionary describing the components of the term. These can be found in
        ``formulae.design_matrices().group.terms_info``
    data : (DataFrame, Series, ndarray)
        The term values.
    prior : Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    """

    group_specific = True

    def __init__(self, name, term, data, prior=None):
        self.name = name
        self.data = data
        self.prior = prior
        self.type = term["type"]
        self.groups = term["groups"]
        self.levels = term["full_names"]
        self.grouper = term["Ji"]
        self.predictor = term["Xi"]
        self.group_index = self.invert_dummies(self.grouper)
        self.categorical = False

        # Determine if the expression is categorical
        if self.type == "interaction":
            if any((t["type"] == "categoric" for t in term["terms"].values())):
                self.categorical = True
        else:
            self.categorical = self.type == "categoric"

        # Determine if the term represents cell-means encoding.
        self.is_cell_means = self.categorical and (self.data.sum(1) == 1).all()

        # Used in pymc3 model coords to label coordinates appropiately
        self.pymc_coords = {}
        # Group is always a coordinate added to the model.
        expr, factor = self.name.split("|")
        self.pymc_coords[factor + "_coord_group_factor"] = self.groups

        if self.categorical:
            name = expr + "_coord_group_expr"
            if self.type == "interaction":
                self.pymc_coords[name] = term["levels"]
            elif term["encoding"] == "full":
                self.pymc_coords[name] = term["levels"]
            else:
                self.pymc_coords[name] = term["levels"][1:]

    def invert_dummies(self, dummies):
        """
        For the sake of computational efficiency (i.e., to avoid lots of large matrix
        multiplications in the backends), invert the dummy-coding process and represent full-rank
        dummies as a vector of indices into the coefficients.
        """
        vec = np.zeros(len(dummies), dtype=int)
        for i in range(1, dummies.shape[1]):
            vec[dummies[:, i] == 1] = i
        return vec

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"prior: {self.prior}",
            f"groups: {self.groups}",
            f"type: {self.type}",
            f"shape: {self.data.squeeze().shape}",
            f"categorical: {self.categorical}",
        ]
        if self.categorical:
            args += [f"levels: {self.levels}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()
